Attribute VB_Name = "ImgDecModule"
Option Explicit

' Clickteam Image En-/Decoder
' ---------------------------
' There are the following Versions:
'   Version 1 (Klik 'n' Play / Multimedia Fusion / Icons for TGF/CnC/MMFx
'   Version 2 (The Games Factory / Click 'n' Create / MMF Express
' Decoded are:  Version 1 / 2
' Encoded are:  Version 1 (currently without compression - TODO but easy)
'
' Decoded output is Scanline-friendly.
' Input to encode must have Scanlines.

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" _
    (ByRef hpvDest As Any, ByRef hpvSource As Any, ByVal cbCopy As Long)

Public Type IMG_HEADER
    SizeX As Integer
    SizeY As Integer
    HotSpotX As Integer
    HotSpotY As Integer
    ActPntX As Integer
    ActPntY As Integer
    ImgMode As Byte
    TransCol As Long    ' MMF2 only
End Type
Private Type IMG_HEADER_INTERNAL
    SizeX As Integer
    SizeY As Integer
    BitCount As Byte
    ImgFormat As Byte
    ImgMode As Byte
    TransCol As Long    ' MMF2 only
End Type
Public Type IMGDEC_STATS
    PixCount As Long
    PixOpaque As Long
    PixTransp As Long
End Type

Private Type LINE_ENTRY
    ImgPos As Long
    LinePos As Long
    ColCount As Long
    ColNum() As Byte
End Type
Private Type LINE_TOC
    LineCount As Long
    Unknown As Long
    Lines() As LINE_ENTRY
End Type

Private Type COL_CLT_ENTRY
    ColCount As Byte
    ColVal As Long
End Type
Private Type COLOR_COLLECTION
    CltCount As Long
    ColClt() As COL_CLT_ENTRY
End Type

Public NULL_COL_8B As Long
Public NULL_COL_15B As Long
Public NULL_COL_16B As Long
Public NULL_COL_24B As Long

Private ImgSize As Long
Private ImgData() As Byte

Public ImgDecStat As IMGDEC_STATS
Private ImgHead As IMG_HEADER_INTERNAL
Private COL_SIZE As Long
Private OpaqueMode As Boolean

Private PicSize As Long
Private PicData() As Byte

Private MsgForm As Form

Public Sub LoadImgData(ByVal DataSize As Long, ByRef DataArr() As Byte, _
                        ByVal ImgSizeX As Integer, ByVal ImgSizeY As Integer, _
                        ByVal ImgBitCount As Byte, ByVal ImgFormat As Byte, _
                        ByVal ImgMode As Byte, ByVal ImgTransCol As Long)

    ImgSize = DataSize
    ImgData() = DataArr()
    With ImgHead
        .SizeX = ImgSizeX
        .SizeY = ImgSizeY
        .BitCount = ImgBitCount
        .ImgFormat = ImgFormat
        .ImgMode = ImgMode
        .TransCol = ImgTransCol
        COL_SIZE = Int((.BitCount + 7) / 8)
    End With

End Sub

Public Sub ImgDecSetOptions(ByVal OpaqueColors As Boolean)

    OpaqueMode = OpaqueColors

End Sub

Public Function DecodeImgV1() As Byte

    Dim CurPos As Long
    Dim CurCol As Integer
    Dim ColCnt As Integer
    Dim NullColor As Long
    Dim NullChk As Long
    Dim PixPos As Long
    Dim CurX As Long
    Dim LineSize As Long
    Dim PicMode As Boolean
    Dim PixOp As Long
    Dim PixTr As Long
    Dim TempCol As Long
    
    On Error GoTo DecodingError
    
    Select Case COL_SIZE
    Case &H1
        If Not OpaqueMode Then
            'If ImgHead.ImgFormat < &H20 Then
                NullColor = NULL_COL_8B
                NullChk = NullColor
            'Else 'If ImgHead.ImgFormat = &H20 Then
            '    For NullChk = &H0 To &HFF
            '        Call CopyMemory(TempCol, Pal8B(NullChk), &H4)
            '        If TempCol = ImgHead.TransCol Then Exit For
            '    Next NullChk
            '    If NullChk > &HFF Then NullChk = &H0
            'End If
        Else
            NullColor = &H0
            NullChk = &HFFFFFFFF
        End If
    Case &H2
        If Not OpaqueMode Then
            If ImgHead.BitCount = 16 Then
                NullColor = NULL_COL_16B
                TempCol = ((ImgHead.TransCol And &HF80000) / &H80000 * &H1&) Or _
                            ((ImgHead.TransCol And &HFC00&) / &H400& * &H20&) Or _
                            ((ImgHead.TransCol And &HF8&) / &H8& * &H800&)
            Else
                NullColor = NULL_COL_15B
                TempCol = ((ImgHead.TransCol And &HF80000) / &H80000 * &H1&) Or _
                            ((ImgHead.TransCol And &HF800&) / &H800& * &H20&) Or _
                            ((ImgHead.TransCol And &HF8&) / &H8& * &H400&)
            End If
            If ImgHead.ImgFormat < &H20 Then
                NullChk = &H0
            Else 'If ImgHead.ImgFormat = &H20 Then
                NullChk = TempCol
            End If
        Else
            NullColor = &H0
            NullChk = &HFFFFFFFF
        End If
    Case &H3
        If Not OpaqueMode Then
            NullColor = NULL_COL_24B
            If ImgHead.ImgFormat < &H20 Then
                NullChk = &H0
            Else 'If ImgHead.ImgFormat = &H20 Then
                NullChk = ImgHead.TransCol
            End If
        Else
            NullColor = &H0
            NullChk = &HFFFFFFFF
        End If
    End Select
    CurX = CLng(ImgHead.SizeX) * COL_SIZE
    ' KnP's Scanlines have a size of 2*x
    ' Windows' Scanlines have a size of 4*x
    LineSize = Int((CurX + &H1) / &H2) * &H2
    PicSize = Int((CurX + &H3) / &H4) * &H4
    PicSize = PicSize * ImgHead.SizeY
    ReDim PicData(&H0 To PicSize - 1)
    
    If ImgHead.ImgFormat < &H20 Then
        PicMode = CBool(ImgHead.ImgMode And &H1)
    Else
        PicMode = True
    End If
    
RetryDecoding:
    CurPos = &H0
    PixPos = &H0
    PixOp = &H0
    PixTr = &H0
    CurX = &H0
    Do While CurPos < ImgSize
        ColCnt = ImgData(CurPos)
        If PicMode Then
            CurPos = CurPos + &H1
        Else
            ColCnt = &H80 Or ImgHead.SizeX
        End If
        'If (ColCnt And &H7F) = &H0 Then Exit Do
        
        If Not CBool(ColCnt And &H80) Then
            For CurCol = &H0 To (ColCnt And &H7F) - 1
                If OpaqueMode Then
                    ' it's faster
                    Call CopyMemory(PicData(PixPos), ImgData(CurPos), COL_SIZE)
                Else
                    Call CopyMemory(TempCol, ImgData(CurPos), COL_SIZE)
                    If TempCol = NullChk Then
                        TempCol = NullColor
                        PixTr = PixTr + &H1
                    Else
                        PixOp = PixOp + &H1
                    End If
                    Call CopyMemory(PicData(PixPos), TempCol, COL_SIZE)
                End If
                PixPos = PixPos + COL_SIZE
                CurX = CurX + COL_SIZE
                
                If CurX >= LineSize Then
                    PixPos = Int((PixPos + &H3) / &H4) * &H4
                    CurX = CurX - LineSize
                End If
            Next CurCol
            CurPos = CurPos + COL_SIZE
        Else
            For CurCol = &H0 To (ColCnt And &H7F) - 1
                If OpaqueMode Then
                    Call CopyMemory(PicData(PixPos), ImgData(CurPos), COL_SIZE)
                Else
                    Call CopyMemory(TempCol, ImgData(CurPos), COL_SIZE)
                    If TempCol = NullChk Then
                        TempCol = NullColor
                        PixTr = PixTr + &H1
                    Else
                        PixOp = PixOp + &H1
                    End If
                    Call CopyMemory(PicData(PixPos), TempCol, COL_SIZE)
                End If
                PixPos = PixPos + COL_SIZE
                CurX = CurX + COL_SIZE
                CurPos = CurPos + COL_SIZE
                
                If CurX >= LineSize Then
                    PixPos = Int((PixPos + &H3) / &H4) * &H4
                    CurX = CurX - LineSize
                End If
            Next CurCol
        End If
    Loop
    
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV1 = &H0
    
    Exit Function

DecodingError:

    'Stop
    'Resume
    If PicMode And False Then
        Call SendMsg("Decoding Error - use Small-Picture-Mode")
        PicMode = False
        Resume RetryDecoding
    End If
    
    Call SendMsg("Decoding Error")
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV1 = &HFF
    
    Exit Function

End Function

Public Function DecodeImgV2_Old() As Byte

    Dim ImgLines As LINE_TOC
    Dim ImgCols As COLOR_COLLECTION
    Dim ImgMode As Boolean
    Dim CurPos As Long
    Dim CurLin As Long
    Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim ColPos As Long
    Dim CurX As Integer
    Dim CurY As Integer
    Dim TFlag As Boolean
    Dim NullColor As Long
    Dim PixPos As Long
    Dim PixOp As Long
    Dim PixTr As Long
    
    On Error GoTo DecodingError
    
    Select Case COL_SIZE
    Case &H1
        If Not OpaqueMode Then
            NullColor = NULL_COL_8B
        Else
            NullColor = &H0
        End If
    Case &H2
        If Not OpaqueMode Then
            If ImgHead.BitCount = 16 Then
                NullColor = NULL_COL_16B
            Else
                NullColor = NULL_COL_15B
            End If
        Else
            NullColor = &H0
        End If
    Case &H3
        If Not OpaqueMode Then
            NullColor = NULL_COL_24B
        Else
            NullColor = &H0
        End If
    End Select
    PicSize = CLng(ImgHead.SizeX) * COL_SIZE
    PicSize = Int((PicSize + &H3) / &H4) * &H4
    PicSize = PicSize * ImgHead.SizeY
    ReDim PicData(&H0 To PicSize - 1)
    
    ImgMode = CBool(ImgHead.ImgMode And &H7)
    CurPos = &H0
    With ImgLines
        Call CopyMemory(.Unknown, ImgData(CurPos + &H0), &H4)
        Call CopyMemory(TempLng, ImgData(CurPos + &H4), &H4)
        .LineCount = Int(TempLng / &H8)
        If .LineCount <> ImgHead.SizeY Then
            Call SendMsg("LineCount != Image Size Y")
        End If
        ReDim .Lines(&H0 To .LineCount - 1)
        CurPos = CurPos + &H4
        
        TempLng = CurPos
        For CurLin = &H0 To .LineCount - 1
            With .Lines(CurLin)
                Call CopyMemory(.ImgPos, ImgData(CurPos + &H0), &H4)
                .ImgPos = TempLng + .ImgPos
                Call CopyMemory(.LinePos, ImgData(CurPos + &H4), &H4)
                .LinePos = TempLng + .LinePos
                CurPos = CurPos + &H8
            End With
        Next CurLin
        
        For CurLin = &H0 To .LineCount - 1
            With .Lines(CurLin)
                If CurPos <> .ImgPos Then
                    Call SendMsg("Line Position Error:" & vbTab & _
                                "Line Pos: " & Format$(.ImgPos) & vbTab & _
                                "Cur Pos: " & Format$(CurPos))
                End If
                CurPos = .ImgPos
                
                TempLng = &H0
                ColPos = 0
                TFlag = True
                Do While ColPos < ImgHead.SizeX
                    ColPos = ColPos + ImgData(CurPos + TempLng)
                    TFlag = Not TFlag
                    TempLng = TempLng + &H1
                Loop
                If Not TFlag Then
                    ColPos = ColPos + ImgData(CurPos + TempLng)
                    If ColPos <> ImgHead.SizeX Then Stop
                    TempLng = TempLng + &H1
                End If
                
                .ColCount = TempLng
                ReDim .ColNum(&H0 To .ColCount - 1)
                Call CopyMemory(.ColNum(&H0), ImgData(CurPos), TempLng)
                CurPos = CurPos + TempLng
            End With
        Next CurLin
    End With
    
    If CurPos <> ImgLines.Lines(&H0).LinePos Then
        Call SendMsg("Mask Position Error:" & vbTab & _
                    "Mask Pos: " & Format$(ImgLines.Lines(&H0).LinePos) & vbTab & _
                    "Cur Pos: " & Format$(CurPos))
        CurPos = ImgLines.Lines(&H0).LinePos
    End If
    
    With ImgCols
        .CltCount = &H0
        TempLng = &H0
        Do
            If CurPos + TempLng >= ImgSize Then Exit Do
            
            If ImgMode Then
                TempByt = ImgData(CurPos + TempLng)
            Else
                TempByt = &H1
            End If
            If TempByt = &H0 Then Exit Do
            
            If Not CBool(TempByt And &H80) Then
                TempByt = &H1
            Else
                TempByt = TempByt And &H7F
                If TempByt = &H0 Then
                    Call SendMsg("Warning! 0-Pix-Count Color!")
                End If
            End If
            If ImgMode Then
                TempLng = TempLng + &H1 + TempByt * COL_SIZE
            Else
                TempLng = TempLng + &H1 * COL_SIZE
            End If
            .CltCount = .CltCount + TempByt
        Loop
        
        If .CltCount > &H0 Then
            ReDim .ColClt(&H0 To .CltCount - 1)
        Else
            ReDim .ColClt(&H0 To &H0)
        End If
        CurLin = &H0
        Do
            If CurPos >= ImgSize Then Exit Do
            
            If ImgMode Then
                TempByt = ImgData(CurPos)
                CurPos = CurPos + &H1
            Else
                TempByt = &H1
            End If
            
            If TempByt = &H0 Then Exit Do
            
            If Not CBool(TempByt And &H80) Then
                With .ColClt(CurLin)
                    .ColCount = TempByt
                    .ColVal = &H0
                    Call CopyMemory(.ColVal, ImgData(CurPos), COL_SIZE)
                    CurLin = CurLin + &H1
                    CurPos = CurPos + COL_SIZE
                End With
            Else
                For ColPos = &H0 To (TempByt And &H7F) - 1
                    With .ColClt(CurLin)
                        .ColCount = &H1
                        .ColVal = &H0
                        Call CopyMemory(.ColVal, ImgData(CurPos), COL_SIZE)
                        CurLin = CurLin + &H1
                        CurPos = CurPos + COL_SIZE
                    End With
                    If CurPos >= ImgSize Then Exit Do
                Next ColPos
            End If
        Loop
    End With
    
    CurPos = &H0
    TempLng = &H0
    PixPos = &H0
    PixOp = &H0
    PixTr = &H0
    For CurY = &H0 To ImgHead.SizeY - 1
        CurLin = &H0
        TempSht = &H0
        TFlag = True
        For CurX = &H0 To ImgHead.SizeX - 1
            With ImgLines.Lines(CurY)
                Do While TempSht >= .ColNum(CurLin)
                    TempSht = TempSht - .ColNum(CurLin)
                    CurLin = CurLin + &H1
                    TFlag = Not TFlag
                Loop
                TempSht = TempSht + &H1
            End With
            If Not TFlag Then
                If CurPos >= (ImgCols.ColClt(TempLng).ColCount And &H7F) Then
                    CurPos = CurPos - (ImgCols.ColClt(TempLng).ColCount And &H7F)
                    TempLng = TempLng + &H1
                    If TempLng >= ImgCols.CltCount Then
                        TempLng = &H0
                    End If
                End If
                
                With ImgCols.ColClt(TempLng)
                    Call CopyMemory(PicData(PixPos), .ColVal, COL_SIZE)
                    PixPos = PixPos + COL_SIZE
                    CurPos = CurPos + &H1
                    PixOp = PixOp + &H1
                End With
            Else
                Call CopyMemory(PicData(PixPos), NullColor, COL_SIZE)
                PixPos = PixPos + COL_SIZE
                PixTr = PixTr + &H1
            End If
        Next CurX
        PixPos = Int((PixPos + &H3) / &H4) * &H4
    Next CurY
    
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV2_Old = &H0
    
    Exit Function

DecodingError:

    'Stop
    'Resume
    Call SendMsg("Decoding Error")
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV2_Old = &HFF
    
    Exit Function

End Function

Public Function DecodeImgV2() As Byte

    Dim ImgMode As Boolean
    Dim CurPos As Long
    Dim CurOTPos As Long    ' Position in Opaque/Transparent Table
    Dim CurCLPos As Long    ' Position in Color List
    Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim ColPos As Long
    Dim CurX As Integer
    Dim CurY As Integer
    Dim TFlag As Boolean
    Dim ColRemOT As Byte    ' Remaining Colors
    Dim ColRemCL As Byte
    Dim NullColor As Long
    Dim PixPos As Long
    Dim PixOp As Long
    Dim PixTr As Long
    
    On Error GoTo DecodingError
    
    Select Case COL_SIZE
    Case &H1
        If Not OpaqueMode Then
            NullColor = NULL_COL_8B
        Else
            NullColor = &H0
        End If
    Case &H2
        If Not OpaqueMode Then
            If ImgHead.BitCount = 16 Then
                NullColor = NULL_COL_16B
            Else
                NullColor = NULL_COL_15B
            End If
        Else
            NullColor = &H0
        End If
    Case &H3
        If Not OpaqueMode Then
            NullColor = NULL_COL_24B
        Else
            NullColor = &H0
        End If
    End Select
    PicSize = CLng(ImgHead.SizeX) * COL_SIZE
    PicSize = Int((PicSize + &H3) / &H4) * &H4
    PicSize = PicSize * ImgHead.SizeY
    ReDim PicData(&H0 To PicSize - 1)
    
    ImgMode = CBool(ImgHead.ImgMode And &H7)
    CurPos = &H4
    Call CopyMemory(CurOTPos, ImgData(CurPos + &H0), &H4)
    Call CopyMemory(CurCLPos, ImgData(CurPos + &H4), &H4)
    If Int(CurOTPos / &H8) <> ImgHead.SizeY Then
        Call SendMsg("LineCount != Image Size Y")
    End If
    CurOTPos = CurOTPos + &H4
    CurCLPos = CurCLPos + &H4
    
    PixPos = &H0
    PixOp = &H0
    PixTr = &H0
    TFlag = False
    ColRemOT = &H0
    ColRemCL = &H80 ' &H0 would move the file pointer
    For CurY = &H0 To ImgHead.SizeY - 1
        'TFlag = True
        'ColRemOT = ImgData(CurOTPos)
        'CurOTPos = CurOTPos + &H1
        For CurX = &H0 To ImgHead.SizeX - 1
            Do While ColRemOT = &H0
                TFlag = Not TFlag
                ColRemOT = ImgData(CurOTPos)
                CurOTPos = CurOTPos + &H1
            Loop
            If TFlag Then
                Call CopyMemory(PicData(PixPos), NullColor, COL_SIZE)
                PixPos = PixPos + COL_SIZE
                PixTr = PixTr + &H1
            Else
                Do While (ColRemCL And &H7F) = &H0
                    ' Read next Color
                    If Not CBool(ColRemCL And &H80) Then
                        CurCLPos = CurCLPos + COL_SIZE
                    End If
                    If ImgMode Then
                        ColRemCL = ImgData(CurCLPos)
                        CurCLPos = CurCLPos + &H1
                    Else
                        ColRemCL = &H81
                    End If
                Loop
                
                ' Color List: Color Count Byte
                '   Bit 7 = 0 - draw the color n times
                '   Bit 7 = 1 - draw the next x colors 1 time
                Call CopyMemory(PicData(PixPos), ImgData(CurCLPos), COL_SIZE)
                If CBool(ColRemCL And &H80) Then
                    CurCLPos = CurCLPos + COL_SIZE
                End If
                PixPos = PixPos + COL_SIZE
                PixOp = PixOp + &H1
                ColRemCL = ColRemCL - &H1
            End If
            ColRemOT = ColRemOT - &H1
        Next CurX
        PixPos = Int((PixPos + &H3) / &H4) * &H4
    Next CurY
    
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV2 = &H0
    
    Exit Function

DecodingError:

    'Stop
    'Resume
    Call SendMsg("Decoding Error")
    ImgDecStat.PixCount = PixOp + PixTr
    ImgDecStat.PixOpaque = PixOp
    ImgDecStat.PixTransp = PixTr
    DecodeImgV2 = &HFF
    
    Exit Function

End Function

Public Sub GetDecodedPicture(ByRef RetPicSize As Long, RetPicData() As Byte)

    RetPicSize = PicSize
    RetPicData() = PicData()

End Sub

Public Sub ImgDec_SetMsgForm(ByRef FormObj As Form)

    Set MsgForm = FormObj

End Sub

Private Sub SendMsg(ByVal MsgStr As String)

    If MsgForm Is Nothing Then Exit Sub
    
    Call MsgForm.AddMessage(MsgStr)

End Sub

Public Sub EncodePictureV1()

    Dim CurPos As Long
    Dim NewPos As Long
    Dim CurCol As Integer
    Dim ColCnt As Integer
    Dim NullColor As Long
    Dim NullChk As Long
    Dim PixPos As Long
    Dim CurX As Long
    Dim LineSize As Long
    Dim PixCount As Long
    Dim PicMode As Boolean
    Dim TempCol As Long
    
    ' 00 - Draw color n times
    ' 80 - Draw n colors 1 time
    
    CurX = CLng(ImgHead.SizeX) * COL_SIZE
    LineSize = Int((CurX + &H1) / &H2) * &H2
    PixCount = LineSize * ImgHead.SizeY
    PicSize = 2 * PixCount
    ReDim PicData(&H0 To PicSize - 1)
    
    If ImgHead.ImgMode = &H0 Then
        ' no compression
        CurPos = &H0
        PixPos = &H0
        CurX = &H0
        Do While CurPos < ImgSize
            ColCnt = ImgHead.SizeX
            
            For CurCol = &H0 To ColCnt - 1
                Call CopyMemory(PicData(PixPos), ImgData(CurPos), COL_SIZE)
                PixPos = PixPos + COL_SIZE
                CurX = CurX + COL_SIZE
                CurPos = CurPos + COL_SIZE
                
                If CurX >= LineSize Then
                    CurPos = Int((CurPos + &H3) / &H4) * &H4
                    CurX = CurX - LineSize
                End If
            Next CurCol
        Loop
    ElseIf ImgHead.ImgMode = &H1 Then
        ' currently without compression
        CurPos = &H0
        NewPos = &H0
        PixPos = &H0
        CurX = &H0
        Do While CurPos < ImgSize
            'ColCnt = &H7F
            'If ColCnt > PixCount - PixPos Then
            '    ColCnt = PixCount - PixPos
            'End If
            ColCnt = &H80 Or ImgHead.SizeX
            PicData(NewPos) = ColCnt
            NewPos = NewPos + &H1
            
            For CurCol = &H0 To (ColCnt And &H7F) - 1
                Call CopyMemory(PicData(NewPos), ImgData(CurPos), COL_SIZE)
                NewPos = NewPos + COL_SIZE
                CurX = CurX + COL_SIZE
                PixPos = PixPos + &H1
                CurPos = CurPos + COL_SIZE
                
                If CurX >= LineSize Then
                    CurPos = Int((CurPos + &H3) / &H4) * &H4
                    CurX = CurX - LineSize
                End If
            Next CurCol
            'NewPos = NewPos + COL_SIZE
        Loop
    End If
    PicData(NewPos + &H0) = &H0 ' Checksum?
    PicData(NewPos + &H1) = &H0
    PicData(NewPos + &H2) = &H0
    NewPos = NewPos + &H3
    
    PicSize = NewPos
    ReDim Preserve PicData(&H0 To PicSize - 1)

End Sub
